#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from urllib.parse import urlparse


def as_term(value: str) -> str:
    try:
        return split_uri(value)[1]
    except AttributeError:
        return value


def split_uri(uri):
    parsed = urlparse(uri)
    if parsed.fragment:
        fragment = parsed.fragment
    else:
        path_split = parsed.path.split("/")
        fragment = path_split[len(path_split)-1]
    return uri.replace(fragment, ""), fragment
